# osm conflation audit

This website takes a JSON output from [OSM Conflator](https://github.com/mapsme/osm_conflate)
and presents logged-in users an interface for validating each imported point, one-by-one

It records any changes and produces a file that can be later fed back to osm_conflate

## author and license

This is a [fork of a project written by Ilya Zverev](https://github.com/mapsme/cf_audit) for MAPS.ME. Published under Apache License 2.0

## setup

### python dependencies

#### pipenv

you can use [pipenv](https://pipenv.readthedocs.io/en/latest/)

```sh
pipenv install
pipenv shell
./run.py
```

_note - pipenv may not be able to resolve dependencies (this is possibly an issue with a now [deprecated version of the oauth lib](https://stackoverflow.com/a/50452983)) - it is worth trying the suggestions pipenv gives_ (i tend to use `pipenv install --skip-lock`):

```sh
[pipenv.exceptions.ResolutionFailure]: Warning: Your dependencies could not be resolved. You likely have a mismatch in your sub-dependencies.
  First try clearing your dependency cache with $ pipenv lock --clear, then try the original command again.
 Alternatively, you can use $ pipenv install --skip-lock to bypass this mechanism, then run $ pipenv graph to inspect the situation.
  Hint: try $ pipenv lock --pre if it is a pre-release dependency.
```
...all going well, you will see:

```sh
Installing dependencies from Pipfile…
  🐍   ▉▉▉▉▉▉▉▉▉▉▉▉▉▉▉▉▉▉▉▉▉▉▉▉▉▉▉▉▉▉▉▉ 3/3 — 00:00:12
To activate this project's virtualenv, run pipenv shell.
Alternatively, run a command inside the virtualenv with pipenv run.
```
so to start the audit web app, you could do:

```#!/bin/sh
pipenv run ./run.py
```

#### pip

if you don't have pipenv, but do have pip, you can use pip and `requirements.txt`

```sh
pip install -r requirements.txt
```

### osm oauth app registration

you'll have to register your client  app on [OpenStreetMap oauth](https://www.openstreetmap.org/user/)

_make sure to select the `read their user preferences` option_

![make sure to tick the read their user preferences option](img/edit-osm-client-app-oauth.png)

### mapillary client_id registration

as with osm, you can also [register the app and callback url](mapillary) to help with surveying sites in street view perspective

ultimately the callback url should be a public url but during early stages localhost can be used

once registered, you can add the mapillary client_id to `config.py`

### admin record in sql

using sqlite you can obtain the id number of your logged in user and add this to `config.py`

```sh
sqlite3 audit.db
```

```sql
.tables
feature  project  task     user     version

.schema user
CREATE TABLE IF NOT EXISTS "user" ("uid" INTEGER NOT NULL PRIMARY KEY, "admin" INTEGER NOT NULL, "bboxes" TEXT);

select * from user;
9465963|0|51.01375465718821,-1.8999481201171877,52.55715099278441,1.3183593750000002
```

## creating a project

### preceding partner project

the initial conflation [`.json`](https://maps.london.gov.uk/gitea/joelondon/osm-london-charging-stations/src/branch/master/preview.json) and [`.osm` audit](https://maps.london.gov.uk/gitea/joelondon/osm-rapid-charging-stations/src/branch/master/results.osm) files that feed into this are created run in a project [containing the `profile.py` required to output the necessary files](https://maps.london.gov.uk/gitea/joelondon/osm-rapid-charging-stations)

[some](https://lists.openstreetmap.org/pipermail/talk-gb-london/2019-March/thread.html) [help](https://lists.openstreetmap.org/pipermail/talk-gb-london/2019-April/thread.html) was given by osm community members, and the [board of the osm cic](https://osmuk.org/directors/) recommended using the osm conflation tool and audit frontend

### making an audit project

having created conflation `.json` and `.osm` files, and made yourself an admin in your client app db (see section on sqlite above), log in and you'll see the option to create/update a project

_make sure the project short name has no spaces in it as it is used to form the url_

_make sure to select the `Enable validation` option_

![make sure to tick the enable validation option](img/edit-osm-conflation-audit-project.png)

### taking it to the osm community

_the client app should be published online and made available to the osm community for the purposes of validating the data_

by including a `start.sh` it is possible to clone this repo to glitch, providing a secure public url:

#### start.sh example
```sh
# For development use (simple logging, etc):
python run.py
# For production use:
# gunicorn run:app -w 1 --log-file -
```
once you've got a URL, now you're ready to involve the osm community with your data checks, implementing the longer, but more valid, route around [Ilya Zverev's procedure](https://wiki.openstreetmap.org/wiki/OSM_Conflator#Community_Validation) shown below:

> You should not upload any imports to OpenStreetMap without discussing them first. In case of conflated points, you have a GeoJSON file to show, so your fellow mappers can find mistakes in tagging and estimate the geometric precision of new points. You can do even better: ask an opinion for each of the imported points.

![https://wiki.openstreetmap.org/wiki/OSM_Conflator#Community_Validation](img/Conflate_audit_chart.jpg)
